"""Algorithm that creates an averaged Dst.

"""
from __future__ import absolute_import

from .Algorithm import Algorithm
from .AlgorithmException import AlgorithmException
from ..ObservatoryMetadata import ObservatoryMetadata
from .. import TimeseriesUtility
import numpy
import obspy.core


class AverageAlgorithm(Algorithm):
    """Algorithm that creates an average of multiple inputs of same channel (i.e., Dst index).

    Parameters
    ----------
    observatories: array_like
        strings specifying multiple observatories to average
    channel: string
        specifies the channel to be processed.
    location: string
        location code for output
    scales: array_like
        floats that scale the corresponding observatories
    min_count: integer
        minimum number of observatories required for valid average
    min_count_start: UTCDateTime
        the beginning of an interval over which to apply min_count
    min_count_end: UTCDateTime
        the end of an interval over which to apply min_count

    Notes
    -----
    The original version of this algorithm allowed no missing inputs. But,
    allowing missing inputs is problematic as one gets nearer to realtime,
    and input streams have differing latencies. So, min_count was added to
    require a minimum number of observatories (more than 1), and the related
    min_count_start and min_count_end were added to allow min_count to be
    applied to a sub-interval (most likely something lagging realtime by
    a sufficient interval to prevent frequent DC shifts in the output)

    """

    def __init__(
        self,
        observatories=None,
        channel=None,
        location=None,
        scales=None,
        min_count=None,
        min_count_start=None,
        min_count_end=None,
    ):
        Algorithm.__init__(self, inchannels=[channel])
        self._npts = -1
        self._stt = -1
        self._stats = None
        self.scales = scales
        self.observatories = observatories
        self.inchannel = channel
        self.outlocation = location
        self.min_count = min_count
        self.min_count_start = min_count_start
        self.min_count_end = min_count_end
        self.observatoryMetadata = ObservatoryMetadata()

    def can_produce_data(self, starttime, endtime, stream):
        """Can Produce data

        By default require all channels to have data at the same time.

        Parameters
        ----------
        starttime: UTCDateTime
            start time of requested output
        end : UTCDateTime
            end time of requested output
        stream: obspy.core.Stream
            The input stream we want to make certain has data for the algorithm
        """
        return TimeseriesUtility.has_any_channels(
            stream, self.get_required_channels(), starttime, endtime
        )

    def check_stream(self, timeseries):
        """checks a stream to make certain the required data
            exists.

        Parameters
        ----------
        timeseries: obspy.core.Stream
            stream to be checked.
        """

        # A stream produced by EdgeFactory should always pass these checks.

        # must have only one channel for each observatory
        if len(timeseries) != len(self.observatories):
            raise AlgorithmException(
                "Expected data for %d stations, received %d \n"
                "Only 1 channel may be averaged at one time"
                % (len(self.observatories), len(timeseries))
            )

        first = True
        # timeseries starttime and number of samples must match
        for ts in timeseries:
            # grab 1st set of stats to use in output.
            # Its values will be good if checks pass.
            if first:
                self._stats = ts.stats
                self._npts = ts.stats.npts
                self._stt = ts.stats.starttime
                first = False

            if ts.stats.npts != self._npts:
                raise AlgorithmException("Received timeseries have different lengths")

            if ts.stats.starttime != self._stt:
                raise AlgorithmException(
                    "Received timeseries have different starttimes"
                )

    def process(self, timeseries):
        """averages a channel across multiple stations and counts available
        observatories per timestep

        Parameters
        ----------

        Returns
        -------
        out_stream:
            new stream object containing the averaged values and available observatories.
        """
        observatories = self.observatories or [t.stats.station for t in timeseries]

        inchannel = self.inchannel or timeseries[0].stats.channel

        outlocation = self.outlocation or timeseries[0].stats.location

        min_count = self.min_count or len(self.observatories)
        min_count_start = self.min_count_start or timeseries[0].stats.starttime
        min_count_end = self.min_count_end or timeseries[0].stats.endtime

        scale_values = self.scales or ([1] * len(timeseries))
        lat_corr = {}
        i = 0
        for obs in observatories:
            new_obs = {str(obs): scale_values[i]}
            lat_corr.update(new_obs)
            i += 1

        # Run checks on input timeseries
        self.check_stream(timeseries)

        # initialize array for data to be appended
        combined = []
        # loop over stations
        for obsy in observatories:

            # lookup latitude correction factor, default = 1.0
            if obsy in lat_corr:
                latcorr = lat_corr[obsy]

            # create array of data for each station
            # and take into account correction factor
            ts = timeseries.select(station=obsy)[0]
            combined.append(ts.data * latcorr)

        # calculate counts
        count_data = numpy.count_nonzero(
            ~numpy.isnan(numpy.column_stack(timeseries)), axis=1
        )

        # calculate averages
        average_data = numpy.nansum(combined, axis=0) / count_data

        # apply min_count
        average_data[count_data < min_count] = numpy.nan

        # apply min_count_start and min_count_end
        # NOTE: the logic here is not very intuitive, but it works as intended
        utc_datetimes = timeseries[0].times(type="utcdatetime")
        average_data[
            (count_data < timeseries.count()) & (utc_datetimes > min_count_end)
        ] = numpy.nan
        average_data[
            (count_data < timeseries.count()) & (utc_datetimes < min_count_start)
        ] = numpy.nan

        # create first output trace metadata
        average_stats = obspy.core.Stats()
        average_stats.station = "USGS"
        average_stats.channel = inchannel
        average_stats.network = "NT"
        average_stats.location = outlocation
        average_stats.starttime = timeseries[0].stats.starttime
        average_stats.npts = timeseries[0].stats.npts
        average_stats.delta = timeseries[0].stats.delta
        # create second output trace metadata
        count_stats = average_stats.copy()
        count_stats.channel = average_stats.channel + "_count"

        # return stream
        stream = obspy.core.Stream(
            (
                obspy.core.Trace(average_data, average_stats),
                obspy.core.Trace(count_data, count_stats),
            )
        )
        return stream

    @classmethod
    def add_arguments(cls, parser):
        """Add command line arguments to argparse parser.

        Parameters
        ----------
        parser: ArgumentParser
            command line argument parser
        """
        parser.add_argument(
            "--average-observatory-scale",
            default=None,
            help="Scale factor for observatories specified with "
            + "--observatory argument",
            nargs="*",
            type=float,
        )
        parser.add_argument(
            "--average-min-count",
            default=None,
            help="Minimum number of inputs required to calculate average",
            type=int,
            metavar="N",
        )
        parser.add_argument(
            "--average-min-count-start",
            default=None,
            help="First UTC date time to apply min_count to YYYY-MM-DDTHH:MM:SS",
            metavar="ISO8601",
            type=obspy.core.UTCDateTime,
        )
        parser.add_argument(
            "--average-min-count-end",
            default=None,
            help="Last UTC date time to apply min_count to YYYY-MM-DDTHH:MM:SS",
            metavar="ISO8601",
            type=obspy.core.UTCDateTime,
        )

    def configure(self, arguments):
        """Configure algorithm using comand line arguments.

        Parameters
        ----------
        arguments: Namespace
            parsed command line arguments
        """
        Algorithm.configure(self, arguments)

        if self._inchannels:
            if len(self._inchannels) > 1:
                raise AlgorithmException("Only 1 input channel can be specified")
            self.inchannel = self._inchannels[0]

        self.observatories = arguments.observatory
        self.scales = arguments.average_observatory_scale
        if self.scales:
            if len(self.observatories) != len(self.scales):
                raise AlgorithmException(
                    "Mismatch between observatories and scale factors"
                )

        self.outlocation = arguments.outlocationcode or arguments.locationcode

        self.min_count = arguments.average_min_count
        self.min_count_start = arguments.average_min_count_start
        self.min_count_end = arguments.average_min_count_end

        if arguments.realtime and not self.min_count_end:
            # if the realtime parameter is defined in arguments (i.e., it is
            # set via the controller), but min_count_end is not, then assume
            # the most recent interval should NOT allow fewer than the full
            # complement of inputs in order to calculate an average (previous
            # "update" intervals will allow fewer than the full complement)
            delta = TimeseriesUtility.get_delta_from_interval(arguments.interval)
            self.min_count_end = arguments.starttime - delta
